self.__precacheManifest = (self.__precacheManifest || []).concat([
  {
    "revision": "d8cc955efa2931dc8e81a343281c1ada",
    "url": "/v1/admin/index.html"
  },
  {
    "revision": "af05ef40279e1d1dc1a6",
    "url": "/v1/admin/static/css/2.1418c1cd.chunk.css"
  },
  {
    "revision": "af05ef40279e1d1dc1a6",
    "url": "/v1/admin/static/js/2.58442cae.chunk.js"
  },
  {
    "revision": "cfe66e252a9e2a8e0fb3c92885e49213",
    "url": "/v1/admin/static/js/2.58442cae.chunk.js.LICENSE.txt"
  },
  {
    "revision": "ba9f439c1312d36f589a",
    "url": "/v1/admin/static/js/main.17810db3.chunk.js"
  },
  {
    "revision": "e033fbbccf592b449b09",
    "url": "/v1/admin/static/js/runtime-main.d9bbd7d2.js"
  },
  {
    "revision": "448c34a56d699c29117adc64c43affeb",
    "url": "/v1/admin/static/media/glyphicons-halflings-regular.448c34a5.woff2"
  },
  {
    "revision": "89889688147bd7575d6327160d64e760",
    "url": "/v1/admin/static/media/glyphicons-halflings-regular.89889688.svg"
  },
  {
    "revision": "e18bbf611f2a2e43afc071aa2f4e1512",
    "url": "/v1/admin/static/media/glyphicons-halflings-regular.e18bbf61.ttf"
  },
  {
    "revision": "f4769f9bdb7466be65088239c12046d1",
    "url": "/v1/admin/static/media/glyphicons-halflings-regular.f4769f9b.eot"
  },
  {
    "revision": "fa2772327f55d8198301fdb8bcfc8158",
    "url": "/v1/admin/static/media/glyphicons-halflings-regular.fa277232.woff"
  }
]);